--- src/Fl_Scroll.cxx	2011-07-19 10:49:30.000000000 +0600
+++ src/Fl_Scroll.cxx	2014-06-29 18:32:09.712351411 +0600
@@ -345,8 +345,71 @@
     scrollbar.position(al?X:X+W-scrollbar.w(), (at&&pad)?Y+hscrollbar.h():Y);
     hscrollbar.position((al&&pad)?X+scrollbar.w():X, at?Y:Y+H-hscrollbar.h());
   } else {
-    // FIXME recalculation of scrollbars needs to be moved out fo "draw()" (STR #1895)
-    redraw(); // need full recalculation of scrollbars
+   // NOTE recalculation of scrollbars needs to be moved out fo "draw()" (STR #1895)
+   // FIXME the recalculations for position of contents is working for align scrolls: FL_ALIGN_RIGHT | FL_ALIGN_BOTTOM only!
+   children_w += xposition() + Fl::box_dx(box()); // full width of all children
+   children_h += yposition() + Fl::box_dy(box()); // full height of all children
+   bool hide_hscroll = false, hide_scroll = false;
+   bool is_hscroll = hscrollbar.visible() && hscrollbar.maximum() > 0.0;
+   bool is_scroll = scrollbar.visible() && scrollbar.maximum() > 0.0;
+   bool hscroll_end = is_hscroll && int(hscrollbar.maximum()) == xposition();
+   bool scroll_end = is_scroll && int(scrollbar.maximum()) == yposition();
+   int new_x = xposition(), new_y = yposition();
+   
+   if ((is_hscroll || is_scroll) && (dw > 0 || dh > 0) && W >= children_w && H >= children_h) {
+    hide_hscroll = hide_scroll = true; // hide both scrollbars
+   } else {
+    if (dw > 0 && is_scroll && W-scrollbar.w() >= children_w) {
+     hide_hscroll = true; // hide the hscrollbar
+    } else if (dw > 0 && !is_scroll && W >= children_w) {
+     hide_hscroll = true; // hide the hscrollbar
+    }
+    if (dh > 0 && is_hscroll && H-hscrollbar.h() >= children_h) {
+     hide_scroll = true; // hide the scrollbar
+    } else if (dh > 0 && !is_hscroll && H >= children_h) {
+     hide_scroll = true; // hide the scrollbar
+    }
+   }
+   
+   if (dw > 0 && hscroll_end) {
+    new_x = xposition()+(-dw)>0?xposition()+(-dw):0;
+   }
+   
+   if (dh > 0 && scroll_end) {
+    new_y = yposition()+(-dh)>0?yposition()+(-dh):0;
+   }
+   
+   if (hide_hscroll && is_hscroll) {
+    if (!hide_scroll && is_scroll && double(new_y+hscrollbar.h()) > scrollbar.maximum()) {
+     new_y -= hscrollbar.h();
+    }
+   }
+   
+   if (hide_scroll && is_scroll) {
+    if (!hide_hscroll && is_hscroll && double(new_x+scrollbar.w()) > hscrollbar.maximum()) {
+     new_x -= scrollbar.w();
+    }
+   }
+   
+   if (new_x < 0 || (hide_hscroll && new_x+children_w > W)) {
+    new_x = 0;
+   }
+   else if (!hide_hscroll && double(new_x) > hscrollbar.maximum()) {
+    new_x = int(hscrollbar.maximum());
+   }
+   
+   if (new_y < 0 || (hide_scroll && new_y+children_h > H)) {
+    new_y = 0;
+   }
+   else if (!hide_scroll && double(new_y) > scrollbar.maximum()) {
+    new_y = int(scrollbar.maximum());
+   }
+   
+   if (new_x != xposition() || new_y != yposition()) {
+    scroll_to(new_x, new_y);
+   }
+   
+   redraw();
   }
 }
 
